"""
@brief   Module to summarise a video into N frames. There are a lot of methods
         for video summarisation, and a lot of repositories in GitHub, but
         none of them seems to work out of the box. This module contains a 
         simple way of doing it.
@author  Luis C. Garcia Peraza Herrera (luiscarlos.gph@gmail.com).
@date    18 May 2022.
"""

import argparse
import numpy as np
import cv2
import tqdm
import math
import skimage.measure
import scipy
import scipy.spatial.distance
import time
import faiss
import seaborn as sns

# My imports
import videosum


class VideoSummariser():
    def __init__(self, algo, number_of_frames: int = 100, 
            width: int = 1920, height: int = 1080, fps=None,
            time_segmentation=False, segbar_height=32, time_smoothing=0.,
            compute_fid=False):
        """
        @brief   Summarise a video into a collage.
        @details The frames in the collage are evenly taken from the video. 
                 No fancy stuff.

        @param[in]  algo               Algorithm to select key frames.
        @param[in]  number_of_frames   Number of frames of the video you want
                                       to see in the collage.
        @param[in]  width              Width of the summary collage.
        @param[in]  height             Height of the summary collage.
        @param[in]  time_segmentation  Set to True to show a time segmentation
                                       under the video collage.
        @param[in]  segbar_height      Height in pixels of the time
                                       segmentation bar.
        @param[in]  compute_fid        Set it to True if you want a report on
                                       the FID of the summary to the whole
                                       video.
        """
        # Sanity checks
        assert(algo in VideoSummariser.ALGOS)
        assert(number_of_frames > 0)
        assert(width > 0)
        assert(height > 0)
        if (algo == 'time' and time_smoothing > 1e-6):
            raise ValueError('[ERROR] You cannot use time smoothing with the time algorithm.')
        
        # Store attributes
        self.algo = algo
        self.number_of_frames = number_of_frames
        self.width = width
        self.height = height
        self.fps = fps
        self.form_factor = float(self.width) / self.height
        self.time_segmentation = time_segmentation
        self.segbar_height = segbar_height
        self.time_smoothing = time_smoothing
        self.compute_fid = compute_fid

        # Compute the width and height of each collage tile
        self.tile_height = self.height
        nframes = VideoSummariser._how_many_rectangles_fit(self.tile_height, 
                                                          self.width, 
                                                          self.height)
        while nframes < self.number_of_frames: 
            self.tile_height -= 1
            nframes = VideoSummariser._how_many_rectangles_fit(self.tile_height, 
                                                               self.width, 
                                                               self.height)
        self.tile_width = int(round(self.tile_height * self.form_factor))

        # Compute how many tiles per row and column
        self.tiles_per_row = self.width // self.tile_width
        self.tiles_per_col = self.height // self.tile_height 

        # Initialise the array that holds the label of each frame
        self.labels_ = None

        # Initialise the array that holds the indices of the key frames
        self.indices_ = None
 
    @staticmethod
    def _how_many_rectangles_fit(tile_height, width, height):
        """
        @brief Given a certain tile height, this method computes how many
               of them fit in a collage of a given size. We assume that the
               form factor of the tiles in the collage has to be the same of 
               the collage itself.
        @param[in]  tile_height  Tile height.
        @param[in]  width        Width of the collage.
        @param[in]  height       Height of the collage.
        """
        # Compute form factor
        ff = float(width) / height

        # Compute rectangle height 
        tile_width = int(round(tile_height * ff))

        # Compute how many rectangles fit inside the collage
        tiles_per_row = width // tile_width 
        tiles_per_col = height // tile_height

        return tiles_per_row * tiles_per_col 
    
    def _insert_frame(self, im, i, j):
        """@brief Insert image into the collage."""

        # Resize frame to the right tile size
        im_resized = cv2.resize(im, (self.tile_width, self.tile_height), 
                                interpolation = cv2.INTER_LANCZOS4)

        # Insert image within the collage
        y_start = i * self.tile_height
        y_end = y_start + im_resized.shape[0] 
        x_start = j * self.tile_width
        x_end = x_start + im_resized.shape[1] 
        self.collage[y_start:y_end, x_start:x_end] = im_resized
    
    @staticmethod
    def transition_indices(labels):
        """
        @brief This method provides the indices of the boundary frames.

        @details  After the summarisation, the video frames are clustered into
                  classes. This means that there will be a video frame that
                  belongs to class X followed by another frame that belongs to 
                  class Y. This method detects this transitions and returns a
                  list of the Y frames, i.e. the first frames after a class
                  transition.
        @param[in]  labels  Pass the self.labels_ produced after calling
                            the summarise() method.
        @returns a list of indices.
        """
        transition_frames = []

        # Loop over the labels of all the video frames
        prev_class = None
        for idx, l in enumerate(labels):
            if l != prev_class:
                transition_frames.append(idx)
                prev_class = l

        return transition_frames
    
    @staticmethod
    def frame_distance_matrix(n: int):
        """
        @brief Compute the normalised distance matrix of rows and columns. 
        @details The distance matrix of rows and columns is (assuming only four
                 frames):

                    0 1 2 3
                    1 0 1 2
                    2 1 0 1
                    3 2 1 0
                  
                 The normalised version is simply a minmax normalisation of the
                 matrix above.
        """
        # Create lower triangular distance matrix
        dist = np.zeros((n, n), dtype=np.float32)
        for i in range(n):
            for j in range(0, i):
                dist[i, j] = np.abs(i - j)

        # Fill the upper triangular part
        dist = dist + dist.T

        # Normalise matrix to [0, 1]
        return dist / np.max(dist)
    
    # Import the different summarisation methods from their corresponding files
    from ._methods.time import get_key_frames_time
    from ._methods.inception import get_key_frames_inception
    from ._methods.uid import get_key_frames_uid 
    from ._methods.scda import get_key_frames_scda

    def get_key_frames(self, input_path):
        return VideoSummariser.ALGOS[self.algo](self, input_path,
            time_smoothing=self.time_smoothing)

    def check_collage_validity(self, input_path, key_frames):
        """
        @brief Asserts that if there are enough frames to make a collage,
               we should have a full collage.
        @param[in]  input_path  Path to the video.
        @param[in]  key_frames  List of key frames.
        @returns None
        """
        nframes_in_video = videosum.VideoReader.num_frames(input_path, self.fps)
        nframes_in_collage = len(key_frames)
        nframes_requested = self.number_of_frames
        if nframes_in_video > nframes_requested:
            if nframes_in_collage != nframes_requested:
                print("[ERROR] There are {} frames ".format(nframes_in_video) \
                        + "in [{}], but the collage ".format(input_path) \
                        + "only has {} ".format(nframes_in_collage) \
                        + "out of the {} requested.".format(nframes_requested))

    def generate_segbar(self):
        # Create an empty segmentation bar
        segbar_width = self.collage.shape[1]
        segbar = np.full((self.segbar_height, segbar_width, 3), 255, 
            dtype=np.uint8)

        # Create the colour palette, one colour per cluster
        palette = np.array(sns.color_palette("Set3", self.number_of_frames))
        colours = (palette * 255.).astype(np.uint8)

        # Loop over segmentation bar columns
        for c in range(segbar_width):
            # Find the frame corresponding to this vertical bar
            frame_idx = int(round(float(c) * len(self.labels_) / segbar_width))
            frame_idx = np.clip(frame_idx, 0, len(self.labels_) - 1)

            # Find the cluster index of the frame
            cluster_idx = self.labels_[frame_idx]
            
            # Make the line of the colour of the cluster the frame belongs to
            segbar[:, c] = colours[cluster_idx]
        
        # Add the key frame vertical lines to the segmentation bar
        key_frame_line_colour = [0, 0, 0]
        for i in self.indices_:
            # Find the vertical bar corresponding to this frame index
            idx = int(round(i * segbar_width / len(self.labels_)))
            
            # Colour the column corresponding to this key frame
            segbar[:, idx] = key_frame_line_colour

        return segbar

    def summarise(self, input_path):
        """
        @brief Create a collage of the video.  

        @param[in]  input_path   Path to an input video.

        @returns a BGR image (numpy.ndarray) with a collage of the video.
        """
        # Initialise list of frame cluster labels
        self.labels_ = None

        # Get list of key frames
        key_frames = VideoSummariser.ALGOS[self.algo](self, input_path,
            time_smoothing=self.time_smoothing)

        # Ensure that summariser actually filled the labels
        assert(self.labels_ is not None)
        assert(len(self.labels_) == videosum.VideoReader.num_frames(input_path, self.fps))

        # If the video has more frames than those requested, 
        # the collage should have all the frames filled
        self.check_collage_validity(input_path, key_frames)
        
        # Warn the user about the collage having less frames than requested,
        # this is expected if the number of requested frames is higher than
        # the number of frames of the video, but a problem otherwise
        if len(key_frames) < self.number_of_frames:
            print('[WARN] The key frame selection algorithm selected ' \
                + 'less frames than you wanted.')

        # Reset collage image
        self.collage = np.zeros((self.tiles_per_col * self.tile_height, 
                                 self.tiles_per_row * self.tile_width, 3),
                                dtype=np.uint8)
        
        # Initialise collage counters
        i = 0
        j = 0
        for im in key_frames:
            # Insert image in the collage
            self._insert_frame(im, i, j) 

            # Update collage iterators
            j += 1
            if j == self.tiles_per_row:
                j = 0
                i += 1

        # Put dividing lines on the collage
        #for x in range(1, self.tiles_per_row):
        #    self.collage[:, x * self.tile_width] = 255

        if self.time_segmentation:
            # Create the time segmentation bar
            segbar = self.generate_segbar()

            # Glue the time segmentation bar under the collage
            collage_with_seg = np.zeros((self.collage.shape[0] + segbar.shape[0],
                self.collage.shape[1], 3), dtype=self.collage.dtype) 
            collage_with_seg[:self.collage.shape[0], :] = self.collage
            collage_with_seg[self.collage.shape[0]:, :] = segbar
            self.collage = collage_with_seg
        
        # Report the FID between the summary and the whole video if requested
        if self.compute_fid:
            print("[INFO] Summary FID: {}".format(self.fid_storyboard_vs_video(input_path, key_frames)))

        return self.collage

    def fid_storyboard_vs_video(self, input_path, key_frames):
        print("[INFO] Computing FID of the storyboard vs video ...")

        # Initialise InceptionV3 model
        model = videosum.InceptionFeatureExtractor('vector')

        # Compute the multivariate Gaussian of the summary
        story_fv = [model.get_latent_feature_vector(im) for im in key_frames]
        story_mu = np.mean(story_fv, axis=0)
        story_cov = np.cov(story_fv, rowvar=False)

        # Compute the multivariate Gaussian of the whole video 
        # (including the summary of course)
        video_fv = []
        reader = videosum.VideoReader(input_path, sampling_rate=self.fps, 
                                      pix_fmt='rgb24')
        w, h = reader.size
        for raw_frame in tqdm.tqdm(reader):
            im = np.frombuffer(raw_frame, dtype=np.uint8).reshape((h, w, 3))[...,::-1].copy()
            video_fv.append(model.get_latent_feature_vector(im))
        video_mu = np.mean(video_fv, axis=0)
        video_cov = np.cov(video_fv, rowvar=False)

        assert(story_mu.shape == video_mu.shape)
        assert(story_cov.shape == video_cov.shape)

        # Compute 2-Wasserstein distance
        emd = videosum.InceptionFeatureExtractor._calculate_frechet_distance(
            story_mu, story_cov, video_mu, video_cov)

        return emd
    
    # Class attribute: supported key frame selection algorithms
    ALGOS = {
        'time':      get_key_frames_time,
        'inception': get_key_frames_inception,
        'uid' :      get_key_frames_uid,
        'scda':      get_key_frames_scda,
    }


if __name__ == '__main__':
    raise RuntimeError('[ERROR] This is not a Python script.')



"""
@brief   This module provides a class to easily compute the Frechet Inception 
         Distance between to Numpy/OpenCV BGR images.

@author  Luis C. Garcia Peraza Herrera (luiscarlos.gph@gmail.com).
@date    21 Jul 2022.
"""

import numpy as np
import torch
import scipy
import numba

# My imports
import videosum


@numba.jit(nopython=True)
def numba_uid(X, eps=1e-6):
    """
    @brief   The Frechet Inception Distance is computed between two 
             distributions. Here we just have a list of feature vectors. 
             Inspired by FID, in this function we define the 
             Univariate Inception Distance (UID), and compute a distance
             matrix from each vector in X to each other vector in X.

    @details In this function we assume that each feature vector contains
             many samples from the same univariate Gaussian distribution.
             That is, to compare two feature vectors we estimate the mean
             and std of each feature vector and compare the univariate 
             Gaussian distributions: 

             uid = sqrt((u_1 - u_2)**2 + (sigma_1 - sigma_2)**2)
             
    @details a distance matrix.
    """
    uid = np.zeros((X.shape[0], X.shape[0]), dtype=np.float64)
    for i in range(X.shape[0]):
        for j in range(i + 1, X.shape[0]):
            # Compute means
            mu_1 = np.mean(X[i, :])
            mu_2 = np.mean(X[j, :])

            # Compute covariances (variances because it's univariate) 
            C_1 = np.std(X[i, :]) + eps
            C_2 = np.std(X[j, :]) + eps

            # Compute UID
            uid[i, j] = np.sqrt(((mu_1 - mu_2) ** 2) + ((C_1 - C_2) ** 2))
    uid += uid.T
    return uid


class InceptionFeatureExtractor():
    def __init__(self, mode: str, device: str = 'cuda'):
        self.mode = mode
        self.device = device
        
        # Initialise model
        if self.mode == 'vector':
            self.model = videosum.InceptionV3().to(device)
        elif self.mode == 'tensor':
            self.model = videosum.InceptionV3(output_vector=False).to(device)
        else:
            raise ValueError('[ERROR] Output mode not recognized.')
        self.model.eval()
    
    def _get_latent_features(self, im):
        """
        @brief  Given a BGR image, this method computes a 1D feature vector.

        @param[in]  im  Numpy/OpenCV BGR image, shape (H, W, 3),
                        dtype = np.uint8. 
                        Could also be a batch of images with 
                        shape (B, H, W, 3).
        
        @returns a Numpy ndarray of 2048 channels.
        """
        single_image = True if len(im.shape) == 3 else False

        # Convert the image to RGB in range [0, 1]
        im = im[...,::-1].copy().astype(np.float32) / 255.

        # Convert image from Numpy to Pytorch and upload it to GPU
        im = torch.from_numpy(im).to(self.device)

        # Convert image to shape (B, 3, H, W)
        if single_image:
            im = im.permute(2, 0, 1).unsqueeze(0)
        else:
            im = im.permute(0, 3, 1, 2)

        # Run inference on the image and get a vector of shape (2048,)
        with torch.no_grad():
            if single_image:
                output = self.model(im)[0].flatten().cpu().detach().numpy()
            else:
                bs, chan, _, _ = im.shape
                output = torch.squeeze(self.model(im)[0]).cpu().detach().numpy()

        return output

    def get_latent_feature_vector(self, im):
        assert(self.mode == 'vector')
        return self._get_latent_features(im)
    
    def get_latent_feature_tensor(self, im):
        assert(self.mode == 'tensor')
        #return self._get_latent_features(im).reshape((8, 8, 2048))
        return self._get_latent_features(im).reshape((2048, 8, 8))

    @staticmethod
    def frechet_inception_distance(vec1, vec2):
        return InceptionFeatureExtractor._calculate_frechet_distance(
            np.mean(vec1), np.cov(vec1), np.mean(vec2), np.cov(vec2))
    
    #@staticmethod
    #def frechet_inception_distance(mu_sigma_1, mu_sigma_2):
    #    """
    #    @brief Compute the Frechet inception distance between two 
    #           distributions.
    # 
    #    @param[in]  mu_sigma_1  Numpy array of arrays [[means], [covs]] for 
    #                            distribution 1.
    #    @param[in]  mu_sigma_2  Numpy array of arrays [[means], [covs]] for 
    #                            distribution 2.
    #    """
    #    return InceptionFeatureExtractor._calculate_frechet_distance(
    #        mu_sigma_1[0], mu_sigma_1[1], mu_sigma_2[0], mu_sigma_2[0])

    @staticmethod 
    def _calculate_frechet_distance(mu1, sigma1, mu2, sigma2, eps=1e-6):
        """
        @brief Numpy implementation of the Frechet Distance.
               The code of this method was taken from the pytorch-fid package.
        
        @details The Frechet distance between two multivariate Gaussians 
                 X_1 ~ N(mu_1, C_1) and X_2 ~ N(mu_2, C_2) is:

                 d^2 = ||mu_1 - mu_2||^2 + Tr(C_1 + C_2 - 2*sqrt(C_1*C_2))

        Stable version by Dougal J. Sutherland.

        @param[in]  mu1     Numpy array containing the activations of a layer 
                            of the inception net for generated samples.
        @param[in]  mu2     The sample mean over activations, precalculated 
                            on an representative data set.
        @param[in]  sigma1  The covariance matrix over activations for 
                            generated samples.
        @param[in]  sigma2  The covariance matrix over activations, precalculated 
                            on an representative data set.
    
        @returns the Frechet distance.
        """
        mu1 = np.atleast_1d(mu1)
        mu2 = np.atleast_1d(mu2)

        sigma1 = np.atleast_2d(sigma1)
        sigma2 = np.atleast_2d(sigma2)

        assert mu1.shape == mu2.shape, \
            'Training and test mean vectors have different lengths'
        assert sigma1.shape == sigma2.shape, \
            'Training and test covariances have different dimensions'

        diff = mu1 - mu2

        # Product might be almost singular
        covmean, _ = scipy.linalg.sqrtm(sigma1.dot(sigma2), disp=False)
        if not np.isfinite(covmean).all():
            msg = ('fid calculation produces singular product; '
                   'adding %s to diagonal of cov estimates') % eps
            print(msg)
            offset = np.eye(sigma1.shape[0]) * eps
            covmean = linalg.sqrtm((sigma1 + offset).dot(sigma2 + offset))

        # Numerical error might give slight imaginary component
        if np.iscomplexobj(covmean):
            if not np.allclose(np.diagonal(covmean).imag, 0, atol=1e-3):
                m = np.max(np.abs(covmean.imag))
                raise ValueError('Imaginary component {}'.format(m))
            covmean = covmean.real

        tr_covmean = np.trace(covmean)

        return (diff.dot(diff) + np.trace(sigma1)
                + np.trace(sigma2) - 2 * tr_covmean) 


if __name__ == '__main__':
    raise RuntimeError('[ERROR] The FID module is not a script.')



"""
@brief  Video reader class to wrap whatever way we choose to read videos 
        inside the same API.
@author Luis C. Garcia Peraza Herrera (luiscarlos.gph@gmail.com).
@date   22 Nov 2022.
"""
import imageio_ffmpeg


class VideoReader:

    def __init__(self, path: str, sampling_rate: int = None, pix_fmt='rgb24'):
        """
        @param[in]  path           Path to the video.
        @param[in]  sampling_rate  How many frames per second to get from the
                                   video, i.e. sampling fps.
        @param[in]  pix_fmt        Pixel format string compatible with
                                   imageio_ffmpeg.
        """
        # Store parameters
        self.path = path
        self.sampling_rate = sampling_rate
        self.pix_fmt = pix_fmt
        
        # Open video reader
        if self.sampling_rate is None:
            self.reader_ = imageio_ffmpeg.read_frames(self.path,
                input_params=['-hide_banner'],
                pix_fmt=self.pix_fmt)
        else:
            self.reader_ = imageio_ffmpeg.read_frames(self.path, 
                pix_fmt=self.pix_fmt, 
                input_params=['-hide_banner'],
                output_params=[
                    '-filter:v', "fps={}".format(self.sampling_rate),
                ])

        # Get videeo info
        self.meta_ = self.reader_.__next__()

    def __iter__(self):
        return self

    def __next__(self):
        return self.reader_.__next__()

    @property
    def width(self):
        return self.meta_['size'][0]

    @property
    def height(self):
        return self.meta_['size'][1]

    @property
    def size(self):
        return self.meta_['size']

    @property
    def duration(self):
        return self.meta_['duration']

    @property
    def fps(self):
        return self.meta_['fps']

    @staticmethod
    def num_frames(path: str, sampling_rate: int = None):
        """
        @brief Function to count the number of frames in a video.
        @param[in]  path           Path to the video file.
        @param[in]  sampling_rate  At how many fps you want to read the video.
        @returns the number of frames as an integer.
        """
        # Create video reader
        if sampling_rate is None:
            reader = imageio_ffmpeg.read_frames(path, pix_fmt='rgb24', 
                input_params=['-hide_banner'])
        else:
            reader = imageio_ffmpeg.read_frames(path, pix_fmt='rgb24', 
                input_params=['-hide_banner'],
                output_params=['-filter:v', "fps={}".format(sampling_rate)])

        # Read videeo info
        _ = reader.__next__()

        # Count the number of frames
        count = 0
        for _ in reader:
            count += 1

        return count


if __name__ == '__main__':
    raise RuntimeError('[ERROR] The module videosum.videoreader is not a script.')




"""
@brief   Video frame clustering based on Selective Convolutional Descriptor 
         Aggregation.

@details They key frames are selected by unsupervised clustering of 
         latent feature vectors corresponding to the video frames.
         The latent feature vector is obtained as explained in 
         Wei et al. 2017:

         "Selective Convolutional Descriptor Aggregation for
         Fine-Grained Image Retrieval"
                 
         The only difference with the paper is we use InceptionV3
         to retrieve the latent tensor (as opposed to VGG-16).

@author  Luis C. Garcia Peraza Herrera (luiscarlos.gph@gmail.com).
@date    1 Dec 2022.
"""

import numpy as np
import tqdm
import sklearn_extra.cluster
import imageio_ffmpeg
import skimage.measure
import scipy.ndimage
import faiss
import numba

# My imports
import videosum



def get_key_frames_scda(self, input_path, time_smoothing=0.):
    """
    @brief Get a list of key video frames. 

    @param[in]  input_path      Path to the video file.
    @param[in]  time_smoothing  Value in the range [0.0, 1.0]. A higher
                                value makes the clustering closer to that
                                of the 'time' method.

    @returns a list of Numpy/BGR images, range [0.0, 1.0], dtype = np.uint8. 

    """

    @numba.jit(nopython=True)
    def numba_cosine_dist(X):
        dist = np.zeros((X.shape[0], X.shape[0]), dtype=np.float64)
        for i in range(X.shape[0]):
            for j in range(i + 1, X.shape[0]):
                # Compute cosine similarity
                u = X[i, :]
                v = X[j, :]
                u_l2 = np.sqrt(np.sum(u ** 2))
                v_l2 = np.sqrt(np.sum(v ** 2))
                cos_sim = np.dot(u, v) / (u_l2 * v_l2)
                

                # Compute cosine distance
                #cos_dist = 1 - cos_sim
                #dist[i, j] = min(0., max(2., cos_dist))
                dist[i, j] = 1 - cos_sim 
        dist += dist.T
        return dist

    # Initialise video reader
    reader = videosum.VideoReader(input_path, sampling_rate=self.fps,
                                  pix_fmt='rgb24')
    w, h = reader.size

    # Initialise Inception network model
    model = videosum.InceptionFeatureExtractor('tensor')

    # Collect feature vectors for all the frames
    print('[INFO] Collecting feature vectors for all the frames ...')
    latent_vectors = []
    for raw_frame in tqdm.tqdm(reader):
        # Convert video frame into a BGR OpenCV/Numpy image
        im = np.frombuffer(raw_frame, dtype=np.uint8).reshape((h, w, 3))[...,::-1].copy()

        # Compute latent feature vector for this video frame
        tensor = model.get_latent_feature_tensor(im)

        # Sum tensor over channels
        aggregation_map = tensor.sum(axis=0)

        # Compute mask map
        mean = aggregation_map.mean()
        mask_map = np.zeros_like(aggregation_map)
        mask_map[aggregation_map > mean] = 1

        # Connected component analysis
        cc = skimage.measure.label(mask_map) 

        # Largest connected component
        largest = np.argmax(np.bincount(cc.flat)[1:]) + 1
        largest_cc = scipy.ndimage.binary_fill_holes(cc == largest).astype(cc.dtype)

        # Selected descriptor set
        selected_descriptor_set = []
        for i in range(8):
            for j in range(8):
                if largest_cc[i, j]:
                    selected_descriptor_set.append(tensor[:, i, j])
        selected_descriptor_set = np.array(selected_descriptor_set)

        # Compute avg&maxPool SCDA feature vector
        avgpool = np.mean(selected_descriptor_set, axis=0)
        maxpool = np.max(selected_descriptor_set, axis=0)
        scda = np.hstack([avgpool, maxpool])

        # Add feature vector to our list
        latent_vectors.append(scda)
    print('[INFO] Done. Feature vectors computed.')

    # Compute cosine distances
    X = np.array(latent_vectors, dtype=np.float32)
    cos_dist = numba_cosine_dist(X)

    # Minmax normalisation of the l2norm distances
    cos_dist /= np.max(cos_dist)

    # Compute the distance matrix with time smoothing if requested
    fdm = videosum.VideoSummariser.frame_distance_matrix(cos_dist.shape[0])
    dist = (1. - time_smoothing) * cos_dist + time_smoothing * fdm

    # Cluster the feature vectors
    print('[INFO] k-medoids clustering ...')
    kmedoids = sklearn_extra.cluster.KMedoids(n_clusters=self.number_of_frames, 
        metric='precomputed',
        init='k-medoids++',
        random_state=0).fit(dist)
    print('[INFO] k-medoids clustering finished.')
    indices = sorted(kmedoids.medoid_indices_.tolist(), reverse=True)
    self.indices_ = [x for x in indices]
    self.labels_ = kmedoids.labels_

    # Retrieve the video frames corresponding to the cluster means
    print('[INFO] Retrieving key frames ...')
    key_frames = []
    reader = videosum.VideoReader(input_path, sampling_rate=self.fps, 
                                  pix_fmt='rgb24')
    counter = 0
    for raw_frame in tqdm.tqdm(reader):
        if counter in self.indices_:
            # Convert video frame into a BGR OpenCV/Numpy image
            im = np.frombuffer(raw_frame, dtype=np.uint8).reshape((h, w, 3))[...,::-1].copy()

            # Add key frame to list
            key_frames.append(im)
            
        counter += 1
    print('[INFO] Key frames obtained.')

    return key_frames





"""
@brief  Method get_key_frames_inception() of the class VideoSummariser.
@author Luis C. Garcia Peraza Herrera (luiscarlos.gph@gmail.com).
@date   29 Nov 2022
"""
import numpy as np
import tqdm
import faiss
import sklearn_extra.cluster
import imageio_ffmpeg 

# My imports 
import videosum


def get_key_frames_inception(self, input_path, eps=1e-6, time_smoothing=0.):
    """
    @brief Get a list of key video frames. 
    @details They key frames are selected by unsupervised clustering of 
             latent feature vectors corresponding to the video frames.
             The latent feature vector is obtained from Inception V3
             trained on ImageNet. The clustering method used is kmedoids.  

    @param[in]  input_path      Path to the video file.
    @param[in]  time_smoothing  Weight of the time smoothing factor. 
                                Higher values make the clustering closer
                                to the 'time' method. The maximum value
                                is 1.
    
    @returns a list of Numpy/BGR images, range [0.0, 1.0], dtype = np.uint8. 
    """
    latent_vectors = []

    # Initialise video reader
    reader = videosum.VideoReader(input_path, sampling_rate=self.fps, 
                                  pix_fmt='rgb24')
    w, h = reader.size

    # Initialise Inception network model
    model = videosum.InceptionFeatureExtractor('vector')

    # Collect feature vectors for all the frames
    print('[INFO] Collecting feature vectors for all the frames ...')
    for raw_frame in tqdm.tqdm(reader):
        # Convert video frame into a BGR OpenCV/Numpy image
        im = np.frombuffer(raw_frame, dtype=np.uint8).reshape((h, w, 3))[...,::-1].copy()

        # Compute latent feature vector for this video frame
        vec = model.get_latent_feature_vector(im)

        # Add feature vector to our list
        latent_vectors.append(vec)
    print('[INFO] Done. Feature vectors computed.')

    # There is no point to cluster if we have less frames in the video 
    # than the number of frames that fit in the collage
    if len(latent_vectors) < self.number_of_frames:
        # All the frames go in the collage 
        self.indices_ = list(range(len(latent_vectors)))
        self.labels_ = list(range(len(latent_vectors)))
    else:
        # Compute L2 distances
        X = np.array(latent_vectors, dtype=np.float32)
        mt = getattr(faiss, 'METRIC_L2')
        l2norm = np.clip(faiss.pairwise_distances(X, X, mt), 0, None)

        # Minmax normalisation of the distance matrix
        l2norm /= np.max(l2norm)

        # Compute the distance matrix with time smoothing if requested
        fdm = videosum.VideoSummariser.frame_distance_matrix(l2norm.shape[0])
        dist = (1. - time_smoothing) * l2norm + time_smoothing * fdm

        # Cluster the feature vectors
        print('[INFO] k-medoids clustering ...')
        kmedoids = sklearn_extra.cluster.KMedoids(n_clusters=self.number_of_frames,
            metric='precomputed',
            method='pam',
            init='k-medoids++',
            random_state=0).fit(dist)
        self.indices_ = kmedoids.medoid_indices_
        self.labels_ = kmedoids.labels_
        print('[INFO] k-medoids clustering finished.')

    # Retrieve the video frames corresponding to the cluster means
    print('[INFO] Retrieving key frames ...')
    key_frames = []
    reader = videosum.VideoReader(input_path, sampling_rate=self.fps, 
                                  pix_fmt='rgb24')
    counter = 0
    for raw_frame in tqdm.tqdm(reader):
        if counter in self.indices_:
            # Convert video frame into a BGR OpenCV/Numpy image
            im = np.frombuffer(raw_frame, 
                dtype=np.uint8).reshape((h, w, 3))[...,::-1].copy()

            # Add key frame to list
            key_frames.append(im)
            
        counter += 1
    print('[INFO] Key frames obtained.')

    return key_frames


if __name__ == '__main__':
    raise RuntimeError('[ERROR] inception.py cannot be run as a script.')



"""
@brief   Script to record the execution time of each of the summarisation 
         methods and print it in markdown for the README.
@author  Luis C. Garcia Peraza Herrera (luiscarlos.gph@gmail.com).
@date    5 Dec 2022.
"""

import argparse
import imageio_ffmpeg 
import numpy as np
import os
import cv2
import tqdm
import time
import multiprocessing
import datetime

# My imports
import videosum


def help(short_option):
    """
    @returns The string with the help information for each command line option.
    """
    help_msg = {
        '-i': 'Path to the video (required: True)',
    }
    return help_msg[short_option]


def parse_cmdline_params():
    """@returns The argparse args object."""
    # Create command line parser
    parser = argparse.ArgumentParser(description='.')
    parser.add_argument('-i', '--input', required=True, type=str, 
                        help=help('-i'))

    # Read parameters
    args = parser.parse_args()

    return args


def main():
    # Read command line parameters
    args = parse_cmdline_params()
    
    # Check that the input video exists
    if not os.path.isfile(args.input):
        raise ValueError('[ERROR] The input video file does not exist.')

    # Get the number of frames in the video
    fps = 30
    vr = videosum.VideoReader(args.input)
    num_frames = vr.num_frames(args.input, fps)

    # Run all the summarisation methods
    num_collage_images = 16
    width = 1920  
    height = 1080
    iterations = 10
    timings = {}
    for i in range(iterations):
        for m in videosum.VideoSummariser.ALGOS:
            vs = videosum.VideoSummariser(m, num_collage_images, width, height, 
                                          fps=fps)
            tic = time.time() 
            vs.summarise(args.input)
            toc = time.time()
            elapsed = toc - tic
            if m not in timings:
                timings[m] = [elapsed / num_frames]
            else:
                timings[m].append(elapsed / num_frames)

    # Compute the average over iterations
    for m in videosum.VideoSummariser.ALGOS:
        timings[m] = sum(timings[m])/ len(timings[m])

    # Print the summarisation time per frame
    print("\n")
    print('| Method | Summarisation time per frame (s) |')
    print('| ------ | -------------------------------- |')
    for m in videosum.VideoSummariser.ALGOS:
        print("| {:10s} | {:1.3f} |".format(m, timings[m]))

    # One hour video sampled at 1fps example
    print("\n")
    print('| Method | Time for a 1h video sampled at 1fps |')
    print('| ------ | ----------------------------------- |')
    for m in videosum.VideoSummariser.ALGOS:
        td = round(timings[m] * 3600)
        print("| {:10s} | {}s |".format(m, td))


if __name__ == '__main__':
    main()



"""
@brief  Unit tests to check the videosum package.
@author Luis Carlos Garcia Peraza Herrera (luiscarlos.gph@gmail.com).
@date   7 June 2022.
"""
import unittest
import cv2
import numpy as np
import seaborn as sns
import tempfile
import uuid
import imageio_ffmpeg 
import os
import tempfile
import string
import random
#import skimage.util

# My imports
import videosum


def create_toy_video(path, num_colours: int = 12, width: int = 640, 
                     height: int = 480, fps=30, pix_fmt_in='rgb24', 
                     pix_fmt_out='yuv420p'):
    """
    @brief Creates a video that displays 12 colours, one colour per second.
    @returns the path to the toy video.
    """
    # Choose 16 random colours
    palette = np.round(np.array(sns.color_palette("Set3", num_colours)) * 255.).astype(np.uint8)

    # Produce all the frames of the video, one per colour
    frames = []
    for i in range(num_colours):
        for j in range(fps):
            # Create video frame
            im = np.ones((height, width, 3), dtype=np.uint8) * palette[i]
            
            # Add frame to the video
            frames.append(im)
    
    # Save the video to the temporary folder
    writer = imageio_ffmpeg.write_frames(path, (width, height), 
                                         pix_fmt_in='rgb24', 
                                         pix_fmt_out='yuv420p',
                                         fps=fps)

    # Seed the generator
    writer.send(None)  

    # Write video to file
    for frame in frames:
        writer.send(frame)
    writer.close()
     

def random_temp_file_path(length=10):
    """
    @param[in]  length  Length of the filename.
    @returns a path to a temporary file without extension located in /tmp/...
    """
    tmpdir = tempfile.gettempdir()
    letters = string.ascii_lowercase
    fname = ''.join(random.choice(letters) for i in range(length))
    return os.path.join(tmpdir, fname)


class TestVideosum(unittest.TestCase):

    def test_video_reader_1fps(self, duration=12, fps=1, eps=1e-6):
        """
        @brief If the toy video displays 12 solid colour frames at 1fps, 
               we should read 12 frames.
        """
        # Create a dummy video of 12s at 1fps
        path = random_temp_file_path() + '.mp4'
        create_toy_video(path, fps=fps)

        # Check that the video is read as expected
        reader = videosum.VideoReader(path, sampling_rate=1)
        self.assertTrue(np.abs(reader.duration - duration) < eps)
        self.assertTrue(np.abs(reader.fps - fps) < eps)

        # Count the number of frames, should be 12 * 1
        count = 0
        for i in reader:
            count += 1
        self.assertTrue(np.abs(count - duration * fps) < eps)

        # Delete dummy video
        os.unlink(path)

    def test_video_reader_30fps(self, duration=12, fps=30, eps=1e-6):
        """
        @brief We should read 12 * 30 frames from the toy video.
        """
        # Create a dummy video of 12s at 30fps
        path = random_temp_file_path() + '.mp4'
        create_toy_video(path, fps=fps)

        # Check that the video is read as expected
        reader = videosum.VideoReader(path, sampling_rate=fps)
        self.assertTrue(np.abs(reader.duration - duration) < eps)
        self.assertTrue(np.abs(reader.fps - fps) < eps)
        
        # Count the number of frames, should be 16 * 1
        count = 0
        for i in reader:
            count += 1
        self.assertTrue(np.abs(count - duration * fps) < eps)

        # Delete dummy video
        os.unlink(path)

    def test_reading_30fps_as_1fps(self, duration=12, src_fps=30, dst_fps=1, 
                              eps=1e-6):
        """
        @brief Even when the toy video is recorded at 30 fps, if we read it
               at 1fps we should read 12 frames, one per second.
        """
        # Create a dummy video of 12s at 30fps
        path = random_temp_file_path() + '.mp4'
        create_toy_video(path, fps=src_fps)

        # Check that the video is read as expected
        reader = videosum.VideoReader(path, sampling_rate=dst_fps)
        self.assertTrue(np.abs(reader.duration - duration) < eps)
        self.assertTrue(np.abs(reader.fps - dst_fps) < eps)
        
        # Count the number of frames, should be 12 * 1
        count = 0
        for i in reader:
            count += 1
        self.assertTrue(np.abs(count - duration * dst_fps) < eps)

        # Remove temporary video file
        os.unlink(path)

    def test_reading_1fps_as_30fps(self, duration=12, src_fps=1, dst_fps=30,
                                   eps=1e-6):
        """
        @brief Even when the toy video is saved at 1fps we should be able to
               read it at 30fps (with duplicated frames).
        """
        # Create a dummy video of 12s at 1fps
        path = random_temp_file_path() + '.mp4'
        create_toy_video(path, fps=src_fps)

        # Check that the video is read as expected
        reader = videosum.VideoReader(path, sampling_rate=dst_fps)
        self.assertTrue(np.abs(reader.duration - duration) < eps)
        
        # Count the number of frames, should be 12 * 30
        count = 0
        for i in reader:
            count += 1
        self.assertTrue(np.abs(count - duration * dst_fps) < eps)

        # Remove temporary video file
        os.unlink(path)

    def test_same_frames_read_at_different_fps(self, eps=1e-6):
        # Create a dummy video of 12s at 30fps
        path = random_temp_file_path() + '.mp4'
        create_toy_video(path, fps=30)

        # Read frames at 30fps
        frames_30fps = []
        reader_30fps = videosum.VideoReader(path, sampling_rate=30)
        w, h = reader_30fps.size
        for raw_frame in reader_30fps:
            im = np.frombuffer(raw_frame, 
                               dtype=np.uint8).reshape((h, w, 3))[...,::-1].copy()
            frames_30fps.append(im)

        # Read frames at 1fps
        frames_1fps = []
        reader_1fps = videosum.VideoReader(path, sampling_rate=1)
        for raw_frame in reader_1fps:
            im = np.frombuffer(raw_frame, 
                               dtype=np.uint8).reshape((h, w, 3))[...,::-1].copy()
            frames_1fps.append(im)

        # Make sure that for each frame of the 1fps video there are 30 identical
        # frames in the 30fps video
        counter = 0
        for f1 in frames_1fps:
            f1 = f1.astype(np.float32)
            init_counter = counter 
            while counter < init_counter + 30:
                f30 = frames_30fps[counter].astype(np.float32)
                diff = np.abs(f30 - f1).sum()
                self.assertTrue(diff < eps)
                counter += 1

        # Remove temporary video file
        os.unlink(path)
         
    def test_time_summary(self, eps=1e-6):
        """
        @brief Check that the summary for the 'time' method is identical to
               the one created by previous versions of the package (which
               was visually inspected and appeared correct).
        """
        # Create dummy video
        video_path = random_temp_file_path() + '.mp4'
        create_toy_video(video_path, fps=30)

        # Load video
        width = 640
        height = 480
        nframes = 12
        vs = videosum.VideoSummariser('time', nframes, width, height, 
                                      time_segmentation=1, fps=30)

        # Make collage
        new_collage_path = 'test/data/time_dummy.png'
        new_collage = vs.summarise(video_path)
        cv2.imwrite(new_collage_path, new_collage)

        # Compare the collage with the one stored in the test folder
        old_collage_path = 'test/data/test_time.png'
        old_collage = cv2.imread(old_collage_path, cv2.IMREAD_UNCHANGED)
        diff = np.sum(np.abs(old_collage.astype(np.float32) - new_collage.astype(np.float32)))
        self.assertTrue(diff < eps)
        
        # Delete dummy video and new collage
        os.unlink(video_path)
        os.unlink(new_collage_path)

    def test_inception_summary(self, eps=1e-6):
        """
        @brief Check that the summary for the 'inception' method is identical to
               the one created by previous versions of the package (which 
               appeared to be correct on visual inspection).
        """
        # Create dummy video
        video_path = random_temp_file_path() + '.mp4'
        create_toy_video(video_path, fps=30)
        
        # Load video
        width = 640
        height = 480
        nframes = 12
        vs = videosum.VideoSummariser('inception', nframes, width, height, 
                                      time_segmentation=1, fps=30)
        
        # Make collage
        new_collage_path = 'test/data/inception_dummy.png'
        new_collage = vs.summarise(video_path)
        cv2.imwrite(new_collage_path, new_collage)
        
        # Compare the collage with the one stored in the test folder
        old_collage_path = 'test/data/test_inception.png'
        old_collage = cv2.imread(old_collage_path, cv2.IMREAD_UNCHANGED)
        diff = np.sum(np.abs(old_collage.astype(np.float32) - new_collage.astype(np.float32)))
        self.assertTrue(diff < eps)

        # Delete dummy video and new collage
        os.unlink(video_path)
        os.unlink(new_collage_path)

    
    def test_same_inception_collage_at_different_fps(self, eps=1e-6):
        """
        @brief No matter the FPS of the video there are 12 solid colors in the
               frames of the video and they should be present in the collage.
        """
        # Create dummy video
        video_path = random_temp_file_path() + '.mp4'
        create_toy_video(video_path, fps=30)
        
        # Make collage at different fps
        width = 640
        height = 480
        nframes = 12
        vs_30fps = videosum.VideoSummariser('inception', nframes, width, height, 
                                           time_segmentation=0, fps=30)
        vs_60fps = videosum.VideoSummariser('inception', nframes, width, height, 
                                           time_segmentation=0, fps=60)
        collage_30fps = vs_30fps.summarise(video_path)
        collage_60fps = vs_60fps.summarise(video_path)
            
        # Make sure that there is no difference, the key frames should be the same
        diff = np.sum(np.abs(collage_30fps.astype(np.float32) - collage_60fps.astype(np.float32)))
        self.assertTrue(diff < eps)

        # Delete dummy video and new collage
        os.unlink(video_path)

    def test_that_indices_and_labels_are_correctly_generated(self, eps=1e-6):
        """
        @brief self.labels_ and self.indices_ must be generate after the video
               summarisation.
        """
        # Create dummy video
        video_path = random_temp_file_path() + '.mp4'
        create_toy_video(video_path, fps=30)
        
        # Create video summarisers for all the methods
        width = 640
        height = 480
        nframes = 12
        fps = 30
        methods = {}
        for method in videosum.VideoSummariser.ALGOS:
            methods[method] = videosum.VideoSummariser(method, nframes, width, height, 
                                                       time_segmentation=1, fps=fps)

        # Check that self.indices_ and self.labels_ are None as we have not
        # run any summary
        for method in methods:
            self.assertTrue(methods[method].indices_ is None)
            self.assertTrue(methods[method].labels_ is None)

        # Run all the summary methods
        for method in methods:
            # Summarise the video into a storyboard
            methods[method].summarise(video_path)

            # Assert that the number of labels corresponds to the number of 
            # frames clustered
            self.assertTrue(len(methods[method].labels_) == nframes * fps)

            # Assert that we have the indices of each cluster center
            indices = methods[method].indices_
            self.assertTrue(len(indices) == nframes)

            # The toy video displays 12 different colours, therefore the 
            # indices of the cluster centers should be different
            self.assertTrue(np.unique(indices).shape[0] == nframes)

        # Delete dummy video and new collage
        os.unlink(video_path)

    def test_fid_method(self, eps=1e-6):
        """
        @brief The 'fid' method should not produce a result different to that
               of the inception method as the toy video contains 12 different
               colours that should be shown in the collage.
        """
        # Create dummy video
        video_path = random_temp_file_path() + '.mp4'
        create_toy_video(video_path, fps=30)

        # Define collage properties
        width = 640
        height = 480
        nframes = 12
        
        # Produce the collage with the inception method
        inception_vs = videosum.VideoSummariser('inception', nframes, width, height, 
                                                time_segmentation=0, fps=30)
        inception_collage = inception_vs.summarise(video_path)

        # Produce the collage with the the fid method
        fid_vs = videosum.VideoSummariser('fid', nframes, width, height, 
                                          time_segmentation=0, fps=30)
        fid_collage = fid_vs.summarise(video_path)

        # Check that the collages produced by both methods are identical
        diff = np.abs(inception_collage.astype(np.float32) - fid_collage.astype(np.float32))
        self.assertTrue(np.sum(diff) < eps)

        # Delete dummy video and new collage
        os.unlink(video_path)
    
    def test_scda_method(self, eps=1e-6):
        """
        @brief The 'scda' method should produce a collage with 12 colours as
               that produced by the inception method. 
        """
        # Create dummy video
        video_path = random_temp_file_path() + '.mp4'
        create_toy_video(video_path, fps=30)

        # Define collage properties
        width = 640
        height = 480
        nframes = 12
        
        # Produce the collage with the inception method
        inception_vs = videosum.VideoSummariser('inception', nframes, width, height, 
                                                time_segmentation=0, fps=30)
        inception_collage = inception_vs.summarise(video_path)

        # Produce the collage with the the scda method
        scda_vs = videosum.VideoSummariser('scda', nframes, width, height, 
                                          time_segmentation=0, fps=30)
        scda_collage = scda_vs.summarise(video_path)

        # Check that the collages produced by both methods are identical
        diff = np.abs(inception_collage.astype(np.float32) - scda_collage.astype(np.float32))
        self.assertTrue(np.sum(diff) < eps)

        # Delete dummy video
        os.unlink(video_path)

    def test_transition_frames(self):
        """
        @brief If we create a video of 12s at 1fps, displaying a different
               colour for every frame, the key frames are the same as the 
               transition frames.
        """
        # Create dummy video
        video_path = random_temp_file_path() + '.mp4'
        create_toy_video(video_path, fps=1)

        # Define collage properties
        width = 640
        height = 480
        number_of_images = 12
        
        # Summarise video
        vs = videosum.VideoSummariser('inception', number_of_images, width, 
                                      height, time_segmentation=0, fps=1)
        inception_collage = vs.summarise(video_path)

        # Assert that there are 12 key frames
        self.assertTrue(len(vs.labels_) == number_of_images)

        # Assert that there are 12 transition frames
        ti = vs.transition_indices(vs.indices_)
        self.assertEqual(len(ti), number_of_images)

        # The key frames and the transition frames should be the same
        self.assertEqual(set(vs.indices_), set(ti))

        # Delete dummy video
        os.unlink(video_path)


if __name__ == '__main__':
    unittest.main()



"""
@brief   Script that reads the CSV results for FID of storyboard vs video and
         plots the FID score averaged across videos (y-axis) for every 
         method (lines) and number of frames (x-axis).

@author  Luis C. Garcia Peraza Herrera (luiscarlos.gph@gmail.com).
@date    13 Jan 2023.
"""

import argparse
import numpy as np
import matplotlib.pyplot as plt
import seaborn as sns
import pandas as pd
import csv
import os


def help(short_option):
    """
    @returns The string with the help information for each command line option.
    """
    help_msg = {
        '-i': 'Path to the CSV file (required: True)',
        '-o': 'Path to save the output plot (required: True)',
    }
    return help_msg[short_option]


def parse_cmdline_params():
    """@returns The argparse args object."""
    # Create command line parser
    parser = argparse.ArgumentParser(description='Easy-to-use video summarisation.')
    parser.add_argument('-i', '--input', required=True, type=str, 
                        help=help('-i'))
    parser.add_argument('-o', '--output', required=True, type=str, 
                        help=help('-o'))

    # Read parameters
    args = parser.parse_args()

    return args


def validate_cmdline_params(args):
    """
    @brief Input directory must exist and output must not.
    """
    if not os.path.isfile(args.input) and not os.path.isdir(args.input):
        raise RuntimeError('[ERROR] Input file or folder does not exist.')

    if os.path.isfile(args.output):
        raise RuntimeError('[ERROR] Output file already exists.')

    return args


def main():
    # Read command line parameters
    args = parse_cmdline_params()
    validate_cmdline_params(args)
    
    # Raw data dictionary
    data = {
        #'video'  : [],
        'Method' : [],
        'Size'   : [],
        'FID'    : [],
    }
    
    # Read CSV file and build Pandas dataframe
    with open(args.input) as csv_file:
        csv_reader = csv.reader(csv_file, delimiter=',')
        line_count = 0
        for row in csv_reader:
            if line_count > 0:
                #data['video'].append(row[0])
                data['Method'].append(row[1])
                data['Size'].append(float(row[2]))
                data['FID'].append(float(row[3]))
            line_count += 1
    df = pd.DataFrame.from_dict(data)
    
    # Average FID scores over videos
    # TODO

    # Generate FID vs no. images for all the methods
    lp = sns.lineplot(data=df, x='Size', y='FID', hue='Method')
    lp.set(title='Storyboard vs Video')
    fig = lp.get_figure()
    plt.legend(loc='upper right')
    plt.xlim([0, 70])
    plt.ylim([40, 250])
    plt.tight_layout()
    fig.savefig(args.output)

    
if __name__ == '__main__':
    main()



"""
@brief  Summarise video as a storyboard where the frame descriptors are the
        mean and the variance of the InceptionV3 feature vector. The 
        distance metric used for clustering is Wasserstein-2. 

@author Luis C. Garcia Peraza Herrera (luiscarlos.gph@gmail.com).
@date   1 Dec 2022.
"""
import tqdm
import numpy as np
import sklearn_extra.cluster
import imageio_ffmpeg

# My imports
import videosum

def get_key_frames_uid(self, input_path, time_smoothing=0.):
        """
        @brief Get a list of key video frames. 
        @details They key frames are selected by unsupervised clustering of 
                 latent feature vectors corresponding to the video frames.
                 The latent feature vector is obtained from Inception V3
                 trained on ImageNet. The clustering method used is kmedoids.  

        @param[in]  input_path      Path to the video file.
        @param[in]  time_smoothing  Weight of the time smoothing factor. 
                                    Higher values make the clustering closer
                                    to the 'time' method. The maximum value
                                    is 1.
        
        @returns a list of Numpy/BGR images, range [0.0, 1.0], dtype = np.uint8. 
        """
        latent_vectors = []

        # Initialise video reader
        reader = videosum.VideoReader(input_path, sampling_rate=self.fps, 
                                      pix_fmt='rgb24')
        w, h = reader.size

        # Initialise Inception network model
        model = videosum.InceptionFeatureExtractor('vector')

        # Collect feature vectors for all the frames
        print('[INFO] Collecting feature vectors for all the frames ...')
        self.frame_count_ = 0
        for raw_frame in tqdm.tqdm(reader):
            self.frame_count_ += 1

            # Convert video frame into a BGR OpenCV/Numpy image
            im = np.frombuffer(raw_frame, dtype=np.uint8).reshape((h, w, 3))[...,::-1].copy()

            # Compute latent feature vector for this video frame
            vec = model.get_latent_feature_vector(im)

            # Add feature vector to our list
            latent_vectors.append(vec)
        print('[INFO] Done. Feature vectors computed.')

        # Compute distance matrix
        print('[INFO] Computing distance matrix ...')
        X = videosum.numba_uid(np.array(latent_vectors))
        print('[INFO] Done, distance matrix computed.')

        # Minmax normalisation of distance matrix
        X /= np.max(X)

        # Time smoothing 
        fdm = videosum.VideoSummariser.frame_distance_matrix(X.shape[0])
        dist = (1. - time_smoothing) * X + time_smoothing * fdm

        # Cluster the feature vectors using the Frechet Inception Distance 
        print('[INFO] k-medoids clustering ...')
        kmedoids = sklearn_extra.cluster.KMedoids(n_clusters=self.number_of_frames, 
            metric='precomputed',
            method='pam',
            init='k-medoids++',
            random_state=0).fit(dist)
        self.indices_ = kmedoids.medoid_indices_
        self.labels_ = kmedoids.labels_
        print('[INFO] k-medoids clustering finished.')

        # Retrieve the video frames corresponding to the cluster means
        print('[INFO] Retrieving key frames ...')
        key_frames = []
        reader = videosum.VideoReader(input_path, sampling_rate=self.fps, 
                                      pix_fmt='rgb24')
        counter = 0
        for raw_frame in tqdm.tqdm(reader):
            if counter in self.indices_:
                # Convert video frame into a BGR OpenCV/Numpy image
                im = np.frombuffer(raw_frame, dtype=np.uint8).reshape((h, w, 3))[...,::-1].copy()

                # Add key frame to list
                key_frames.append(im)
                
            counter += 1
        print('[INFO] Key frames obtained.')

        return key_frames


if __name__ == '__main__':
    raise RuntimeError('[ERROR] uid.py cannot be executed as a script.')
